/****************************************************************************
 * GP32 mp3play library                                                     *
 * Based on :                                                               *
 *																			*
 * madlld.c -- A simple program decoding an MPEG audio stream to 16-bit		*
 * PCM from stdin to stdout. This program is just a simple sample			*
 * demonstrating how the low-level libmad API can be used.					*
 *--------------------------------------------------------------------------*
 * (c) 2001--2004 Bertrand Petit											*
 *																			*
 * Redistribution and use in source and binary forms, with or without		*
 * modification, are permitted provided that the following conditions		*
 * are met:																	*
 *																			*
 * 1. Redistributions of source code must retain the above copyright		*
 *    notice, this list of conditions and the following disclaimer.			*
 *																			*
 * 2. Redistributions in binary form must reproduce the above				*
 *    copyright notice, this list of conditions and the following			*
 *    disclaimer in the documentation and/or other materials provided		*
 *    with the distribution.												*
 * 																			*
 * 3. Neither the name of the author nor the names of its contributors		*
 *    may be used to endorse or promote products derived from this			*
 *    software without specific prior written permission.					*
 * 																			*
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS''		*
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED		*
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A			*
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR		*
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,				*
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT			*
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF			*
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND		*
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,		*
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT		*
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF		*
 * SUCH DAMAGE.																*
 *																			*
 ****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include <mad.h>

#include "gp32.h"
#include "fileio.h"

struct mad_stream	Stream;
struct mad_frame	Frame;
struct mad_synth	Synth;
mad_timer_t			Timer;

unsigned int	INPUT_BUFFER_SIZE;
unsigned char*	InputBuffer = NULL;

GPFILE *mp3File = NULL;
unsigned int PARTIAL_SIZE;

/****************************************************************************
 * Converts a sample from libmad's fixed point number format to a signed	*
 * short (16 bits).															*
 ****************************************************************************/
inline signed short MadFixedToSshort(mad_fixed_t Fixed)
{
	/* Clipping */
	if(Fixed>=MAD_F_ONE)
		return(SHRT_MAX);
	if(Fixed<=-MAD_F_ONE)
		return(-SHRT_MAX);

	/* Conversion. */
	Fixed=Fixed>>(MAD_F_FRACBITS-15);
	return((signed short)Fixed);
}

void MpegInitDecoder()
{
	/* First the structures used by libmad must be initialized. */
	mad_stream_init(&Stream);
	mad_frame_init(&Frame);
	mad_synth_init(&Synth);
	mad_timer_reset(&Timer);
}

void MpegCloseDecoder()
{
	mad_synth_finish(&Synth);
	mad_frame_finish(&Frame);
	mad_stream_finish(&Stream);
	if (InputBuffer!=NULL)
	{
		free(InputBuffer);
		InputBuffer = NULL;
	}
	if (mp3File!=NULL)
	{
		fclose(mp3File);
		mp3File = NULL;
	}
}

int MpegAudioDecoder(signed short *OutputBuffer, unsigned int OUTPUT_BUFFER_SIZE, unsigned int OutputRate)
{
	signed short		*OutputPtr=OutputBuffer;
	const signed short	*OutputBufferEnd=OutputBuffer+OUTPUT_BUFFER_SIZE;
	unsigned char		*GuardPtr=NULL;
	unsigned int		i,step,end;

	do
	{
		if(Stream.buffer==NULL || Stream.error==MAD_ERROR_BUFLEN)
		{
			size_t			ReadSize,
							Remaining;
			unsigned char	*ReadStart;

			if(Stream.next_frame!=NULL)
			{
				Remaining=Stream.bufend-Stream.next_frame;
				memmove(InputBuffer,Stream.next_frame,Remaining);
				ReadStart=InputBuffer+Remaining;
				ReadSize=INPUT_BUFFER_SIZE-Remaining;
			}
			else
			{
				ReadSize=INPUT_BUFFER_SIZE;
				ReadStart=InputBuffer;
				Remaining=0;
			}

			if(ReadStart>=InputBuffer+INPUT_BUFFER_SIZE)
			{
				GuardPtr=ReadStart+ReadSize;
				memset(GuardPtr,0,MAD_BUFFER_GUARD);
				ReadSize+=MAD_BUFFER_GUARD;
			}

			/* Pipe the new buffer content to libmad's stream decoder
             * facility.
			 */
			mad_stream_buffer(&Stream,InputBuffer,ReadSize+Remaining);
			Stream.error=0;
		}

		if(mad_frame_decode(&Frame,&Stream))
		{
			if(MAD_RECOVERABLE(Stream.error))
			{
				continue;
			}
			else
			{
				if(Stream.error==MAD_ERROR_BUFLEN)
					continue;
				else
					return -1;
			}
		}

		mad_timer_add(&Timer,Frame.header.duration);

		mad_synth_frame(&Synth,&Frame);
		
		step = (Synth.pcm.samplerate<<8)/OutputRate;
		end = Synth.pcm.length<<8;

		if(MAD_NCHANNELS(&Frame.header)==2)
		{
			for(i=0;i<end;i+=step)
			{
				*(OutputPtr++) = MadFixedToSshort(Synth.pcm.samples[0][i>>8]);
				*(OutputPtr++) = MadFixedToSshort(Synth.pcm.samples[1][i>>8]);
			}
		}
		else
		{
			for(i=0;i<end;i+=step)
			{
				signed short Sample = MadFixedToSshort(Synth.pcm.samples[0][i>>8]);
				*(OutputPtr++) = Sample;
				*(OutputPtr++) = Sample;
			}
		}
	} while (OutputPtr<OutputBufferEnd);

	return OutputPtr-OutputBuffer;
}

int MpegLoadFile(const char* name, unsigned int START_SIZE)
{
	int ret = 0;
	unsigned int size;

	PARTIAL_SIZE = 0;

	mp3File = fopen(name,"rb");
	if (mp3File==NULL)
		return -1;

	INPUT_BUFFER_SIZE = mp3File->size;
	InputBuffer = (unsigned char*)malloc(INPUT_BUFFER_SIZE+MAD_BUFFER_GUARD);
	if (START_SIZE==0)
		size = INPUT_BUFFER_SIZE;
	else
	{
		if (INPUT_BUFFER_SIZE<=START_SIZE)
			size = PARTIAL_SIZE = INPUT_BUFFER_SIZE;
		else
			size = PARTIAL_SIZE = START_SIZE;
	}
	if (InputBuffer!=NULL)
	{
		if (fread(InputBuffer,1,size,mp3File) == size)
			ret = size;
		else
		{
			free(InputBuffer);
			fclose(mp3File);
			mp3File = NULL;
			PARTIAL_SIZE = 0;
		}
	}
	if (START_SIZE==0)
	{
		fclose(mp3File);
		mp3File = NULL;
		PARTIAL_SIZE = INPUT_BUFFER_SIZE;
	}
	return ret;
}

int MpegLoadNextChunk(unsigned int CHUNK_SIZE)
{
	if (PARTIAL_SIZE<INPUT_BUFFER_SIZE)
	{
		if (mp3File!=NULL)
		{
			fread(InputBuffer+PARTIAL_SIZE,1,CHUNK_SIZE,mp3File);
			PARTIAL_SIZE+=CHUNK_SIZE;
		}
		else
			return 0;
	}
	if (PARTIAL_SIZE>=INPUT_BUFFER_SIZE)
	{
		if (mp3File!=NULL)
		{
			fclose(mp3File);
			mp3File = NULL;
		}
		return 0;
	}
	return 1;
}
